/*********************************************************/
/* file: command.c                                       */
/*                                                       */
/* Copyright (C) 1994 Karl A Holland                     */
/*                                                       */
/* author: Karl A Holland                                */
/*                                                       */
/* created:  18 april  93                                */
/*                                                       */
/* modified:  7 february 94                              */
/*                                                       */
/* contributing authors:                                 */
/*  None as yet. if you add place your name here or below*/
/*                                                       */
/* description:                                          */
/*   command.c contains all the callbacks for parser     */
/*  mode commands, most of them are dungeon master       */
/*  functions.                                           */
/* LICENSING:                                            */
/*   a full license  is provided in the file LICENSE     */
/* which should have been included with this program when*/
/* it was copied. It may be found in the top directory   */
/* of any NORMAL distribution, named LICENSE             */
/*********************************************************/
#include <command.h>

/****************************************/
/* report_square                        */
/*   report the contents of the square  */
/* that the player is standing in       */
/****************************************/
void
report_square(playerstruct * pl)
{
  object         *temp;
  char            tmpbuf[80];
  write_text(pl, "from top to bottom:");
  for (temp = TOP_MAP_OBJ(pl->obj->mapin, pl->obj->x, pl->obj->y, pl->obj->z);
       temp != NULL; temp = temp->below)
  {
    sprintf(tmpbuf, "TYPE: %s NAME:%s", type_names[temp->type], temp->name);
    write_text(pl, tmpbuf);
  }
}

/******************************************/
/* inventory_report_ob                    */
/*   will dump the data about an object in*/
/* the players inventory. if no arguments */
/* supplied, will assume the first object */
/* in the players inventory               */
/******************************************/
void
inventory_report_ob(playerstruct * pl)
{
  char            temp1[15],
                  temp2[15];
  object         *tmp;
 
  /* scan an check to see if the second argument has been supplied */
  if (sscanf(pl->parsing, "%s %[^\n]", temp1, temp2) == 2)
  {
    /* if second argument supplied then search for it in inventory */
    for (tmp = pl->obj->holding; tmp != NULL && strcmp(tmp->name, temp2);
         tmp = tmp->below);
    if (tmp != NULL)
      report_object(pl, tmp);
    else
      write_text(pl, "no such object in inventory");
  } else
  {
    if ((tmp = pl->obj->holding) != NULL)
      report_object(pl, tmp);
    else
      write_text(pl, "no item in inventory to dump");
  }
  fprintf(stderr, "\n");
}

/**********************************************/
/* inventory_destroy                          */
/*   will destroy ('free') an object in the   */
/* players inventory. if no arguments supplied*/
/* then assumes the first object in the       */
/* players inventory.                         */
/**********************************************/
void
inventory_destroy(playerstruct * pl)
{
  char            temp1[15],
                  temp2[15];
  object         *tmp;
  
  /* scan an check to see if the second argument has been supplied */
  if (sscanf(pl->parsing, "%s %[^\n]", temp1, temp2) == 2)
    {
      /* if so then search for the object in inventory */
      for (tmp = pl->obj->holding; tmp != NULL && strcmp(tmp->name, temp2);
	   tmp = tmp->below);

      /* if the object has been found then destroy it */
      if (tmp != NULL)
	{
	  remove_held(tmp);
	  free_object(tmp);
      } else
	/* otherwise notify */
	write_text(pl, "no such object in inventory");
  } else
    {
      /* is there an item in inventory ?*/
      if ((tmp = pl->obj->holding) != NULL)
	{
	  /* if so then destroy it */
	  remove_held(tmp);
	  free_object(tmp);
      } else
	/* otherwise notify player */
	write_text(pl, "no item in inventory to dump");
    }
}

/*********************************************/
/* dm_goto                                   */
/*   will transport a player to supplied     */
/* coordinates within the same map.          */
/*********************************************/
void
dm_goto(playerstruct * pl)
{
  char            temp[20];
  int             x,
                  y,
                  z;

  /* have 3 arguments after command ? */
  if (sscanf(pl->parsing, "%s %d %d %d", temp, &x, &y, &z) == 4)
  {
    /* pluck the player off the map */
    remove_ob(pl->obj);
    /* slap him down at the specified coordinates */
    insert_ob_in_map(pl->obj->mapin, pl->obj, x, y, z);
  } else
    /* not enough arguments */
    write_text(pl, "bad/insufficient arguments to dm_goto");
}

/*********************************************/
/* dm_place                                  */
/*   will remove the first object in the     */
/* players inventory and insert it into the  */
/* map directly on top of the player         */
/*********************************************/
void
dm_place(playerstruct * pl)
{
  object         *item,
                 *curr;

  /* select the first item in inventory */
  item = pl->obj->holding;

  /* does the player have an object ? */
  if (item == NULL)
  {
    /* no.... tell the player so */
    write_text(pl, "no object to place....");
    return;
  }
  
  /* player has na item so remove it from inventory */
  remove_held(item);

  /* set the direction of the object to face same way as player */
  for (curr = item; curr != NULL; curr = curr->link)
    curr->facing = pl->obj->facing;

  /* insert it into the map */
  insert_ob_in_map(pl->obj->mapin, item, pl->obj->x, pl->obj->y, pl->obj->z);
}

/*********************************************/
/* set_game_speed                            */
/*   will set the time delay factor between  */
/* game turns. if no argument supplied, will */
/* return the current value of the delay     */
/*********************************************/
void
set_game_speed(playerstruct * pl)
{
  char            temp[30];
  int             x;

  if (sscanf(pl->parsing, "%s %d", temp, &x) == 2)
    tmax = x;

  else
  {
    sprintf(temp, "delay = %d", tmax);
    write_text(pl, temp);
  }
}

/*********************************************/
/* say                                       */
/*   given a string on the command line, will*/
/* attempt to find a player nearby and       */
/* display the message on the players text   */
/* area                                      */
/*********************************************/
void
say(playerstruct * pl)
{

  char            temparg[15],
                  msg[250];
  int             xloc,
                  yloc;
  object         *tmp;
  sscanf(pl->parsing, "%s %[^\n]", temparg, msg);
  
  /* iterate through x and y to find a player */
  /* there has got to be a better way....     */
  /* watch this space ....                    */
  for (xloc = pl->obj->x - 2; xloc < pl->obj->x + 2; xloc++)
    for (yloc = pl->obj->y - 2; yloc < pl->obj->y + 2; yloc++)
      if (IN_MAP(pl->obj->mapin, xloc, yloc))
        for (tmp = TOP_MAP_OBJ(pl->obj->mapin, xloc, yloc, 0);
             tmp != NULL; tmp = tmp->below)
          if (IS_PLAYER(tmp) && (tmp != pl->obj))
            write_text(tmp->data.plcomp->pldata, msg);
}


void
list_obs(playerstruct * pl)
{
  write_text(pl, "command 'list` not finished");
}

/**************************************/
/* quit_pl                            */
/*   will notify game that the player */
/* has quit                           */
/**************************************/
void
quit_pl(playerstruct * pl)
{
  write_text(pl, "quitting");
  SET_PL_HAS_EXITED(pl);
}


/**************************************/
/* get_ob                             */
/*   will attempt to retrieve an      */
/* instance of the object whos name   */
/* is supplied by the player, and     */
/* add to the players inventory       */
/**************************************/
void
get_ob(playerstruct * pl)
{
  char            temp1[15],
                  temp2[15];
  object         *temp;
  if (sscanf(pl->parsing, "%s %s", temp1, temp2) == 2)
  {
    temp = get_general_object(temp2);
    if (temp != NULL)
    {
      add_holding(pl->obj, temp);
      add_object(temp);
      pl->top_held = temp;
      SET_PL_ITEM_CHANGED(pl);
      write_text(pl, "obtained");
    } else
      write_text(pl, "object not found");
  } else
    write_text(pl, "bad args to get_ob");
}

/***************************************/
/* print_version                       */
/*   print the current version of the  */
/* game                                */
/***************************************/
void
print_version(playerstruct * pl)
{
  char            temp[80];
  sprintf(temp, "current version is %s", VERSION);
  write_text(pl, temp);
}


/****************************************/
/* pickup_ob                            */
/*   given a name on the command line,  */
/* this subroutine will attempt to find */
/* the object in the stack of objects   */
/* where the player is standing and to  */
/* remove it from the map and add it to */
/* players inventory. If no name given  */
/* it will attempt to remove the topmost*/
/* object in the stack                  */
/****************************************/
void
pickup_ob(playerstruct * pl)
{
  char            temp1[15],
                  temp2[15];
  object         *tmp;

  /* scan and check if a name has been supplied */
  if (sscanf(pl->parsing, "%s %s", temp1, temp2) == 2)
  {
    /* name supplied so ... */
    /* search the stack from top to bottom for the object whose name matches */
    for (tmp = TOP_MAP_OBJ(pl->obj->mapin, pl->obj->x, pl->obj->y, pl->obj->z);
	 tmp != NULL && strcmp(tmp->name, temp2); tmp = tmp->below);
    
    /* is the object found and is the object not the player ?? */
    if (tmp != NULL && (tmp != pl->obj))
      {
	/* yes.. so.... */
	/* remove it from the map */
	remove_ob(tmp);

	/* add to the players inventory */
	add_holding(pl->obj, tmp);
	pl->top_held= tmp;
	SET_PL_ITEM_CHANGED(pl);

    } else
      /* otherwise notify the player */
      write_text(pl, "no such object in stack");
  } else
    {
      /* search the stack for the topmost object that is not the player */
      for (tmp = TOP_MAP_OBJ(pl->obj->mapin,pl->obj->x,pl->obj->y,pl->obj->z);
	   tmp != NULL && tmp == pl->obj; tmp = tmp->below);

      /* if an object is found..  */
      if (tmp != NULL && (tmp != pl->obj))
	{
	  /* then remove it from the mop */
	  remove_ob(tmp);
	  
	  /* add to the players inventory */
	  add_holding(pl->obj, tmp);
	  pl->top_held = tmp;
	  SET_PL_ITEM_CHANGED(pl);
	  
      } else
	/* otherwise notify problem*/
	write_text(pl, "no items to pickup");
    }
}

/******************************************/
/* dm_world_destroy                       */
/*   given a name, this command will try  */
/* to find the object with that name in   */
/* the stack of object that the player is */
/* standing on and to destroy (free) it.  */
/* if no name supplied then it will       */
/* destroy the first object in the stack  */
/* that is not the player                 */
/******************************************/
void
dm_world_destroy(playerstruct * pl)
{
  char            temp1[15],
                  temp2[15];
  object         *tmp;
  
  /* scan and check to see if a name is supplied */
  if (sscanf(pl->parsing, "%s %s", temp1, temp2) == 2)
  {
    /* search for object whose name matches the given one in the stack */
    for (tmp = TOP_MAP_OBJ(pl->obj->mapin, pl->obj->x, pl->obj->y, pl->obj->z);
     tmp != NULL && strcmp(tmp->name, temp2); tmp = tmp->below);
    
    /* trying to destroy a player?? */
    if (tmp != NULL && IS_PLAYER(tmp))
    {
      /* report error */
      write_text(pl, "You may not destroy other players!");
      return;
    } else if (tmp != NULL) /* object found? */
    {
      /* pluck it from map */
      remove_ob(tmp);
      
      /* mark it as freed */
      free_object(tmp);
    } else
      /* othewise no object... so notify */
      write_text(pl, "no such object in stack to destroy");
  } else /* no name given */
  {
    /* search for first object that is not the player that called */
    for (tmp = TOP_MAP_OBJ(pl->obj->mapin, pl->obj->x, pl->obj->y, pl->obj->z);
         tmp != NULL && (tmp == pl->obj); tmp = tmp->below);
    
    if (tmp != NULL && IS_PLAYER(tmp))
    {
      write_text(pl, "You may not destroy other players!");
      return;
    } else if (tmp != NULL)
    {
      remove_ob(tmp);
      free_object(tmp);
    } else
      write_text(pl, "no items to destroy");
  }
}

void
world_report_ob(playerstruct * pl)
{
  char            temp1[15],
                  temp2[15];
  object         *tmp;
  if (sscanf(pl->parsing, "%s %s", temp1, temp2) == 2)
  {
    for (tmp = TOP_MAP_OBJ(pl->obj->mapin, pl->obj->x, pl->obj->y, pl->obj->z);
     tmp != NULL && strcmp(tmp->name, temp2); tmp = tmp->below);
    if (tmp != NULL)
    {
      report_object(pl, tmp);
    } else
      write_text(pl, "no such object in stack");
  } else
  {
    for (tmp = TOP_MAP_OBJ(pl->obj->mapin, pl->obj->x, pl->obj->y, pl->obj->z);
         tmp != NULL && (tmp == pl->obj); tmp = tmp->below);
    if (tmp != NULL)
    {
      report_object(pl, tmp);
    } else
      write_text(pl, "no items to dump");
  }
}

void
save_map(playerstruct * pl)
{
  char            temp1[15],
                  temp2[15],
                  sbuf[100];

  if (sscanf(pl->parsing, "%s %s", temp1, temp2) == 2)
  {
    sprintf(sbuf, "Saving Current Map with NEW NAME = %s", temp2);
    write_text(pl, sbuf);
    write_text(pl, "Current maps name will change too....");
    strcpy(pl->obj->mapin->name, temp2);
  }
  write_text(pl, "starting map save");
  save_map_data(pl->obj->mapin, TO_MAP_DIR);
  write_text(pl, "map save complete!");
}

void
tele_map(playerstruct * pl)
{
  char            temp1[15],
                  temp2[15];
  int             x,
                  y,
                  z,
                  num;
  if ((num = sscanf(pl->parsing, "%s %s %d %d %d", temp1, temp2, &x, &y, &z)) == 2)
  {
    if (send_to_map(pl->obj, temp2, 0, 0, 0))
    {
      write_text(pl, "Long Distance Teleportation Successful!");
      return;
    }
  } else if (num == 5)
  {
    if (send_to_map(pl->obj, temp2, x, y, z))
    {
      write_text(pl, "Teleporation to specified coords successful!");
      return;
    }
  } else
  {
    write_text(pl, "insufficient arguments to tele");
    return;
  }
  write_text(pl, "The map specified does not exist....");
  write_text(pl, "Creating a NEW map with that name...");
  allocate_map(20, 20, temp2);
  if (!send_to_map(pl->obj, temp2, 0, 0, 0))
    write_text(pl, "This operation has failed miserably!");
}

void
macro_put(playerstruct * pl)
{
  char            temp1[15],
                  temp2[15],
                  temp3[15];
  int             k,
                  xloc,
                  yloc,
                  zloc;
  object         *orig = NULL,
                 *tobj,
                *curr;

  zloc = 0;
  if ((k = sscanf(pl->parsing, "%s %s %s", temp1, temp2, temp3)) < 3)
    orig = duplicate_object(pl->top_held);
  else
    orig = get_general_object(temp3);
  if (k < 2)
    strcpy(temp2, "1f");
  k = (temp2[0] - '0');
  if (orig == NULL)
  {
    write_text(pl, "macro could not find specified object!!!");
    return;
  }
  switch (temp2[1])
  {
  case ('f'):
    if ((k < 11) && (k > -1))
    {
      for (xloc = pl->obj->x - k; xloc <= (signed) pl->obj->x + k; xloc++)
        for (yloc = pl->obj->y - k; yloc <= (signed) pl->obj->y + k; yloc++)
          if (IN_MAP(pl->obj->mapin, xloc, yloc))
          {
	    tobj = duplicate_object(orig);
            for (curr = tobj; curr != NULL; curr = curr->link)
              curr->facing = pl->obj->facing;

            insert_ob_in_map(pl->obj->mapin, tobj, xloc, yloc, pl->obj->z);
          }
    }
    break;
  default:
    write_text(pl, " that macro has not been implemented");
    break;
  }
  free_object(orig);
}

void
macro_destroy(playerstruct * pl)
{
  char            temp1[15],
                  temp2[15];
  int             k,
                  xloc,
                  yloc,
                  zloc;
  object         *tmp,
                 *next;

  zloc = 0;
  if ((k = sscanf(pl->parsing, "%s %s", temp1, temp2)) < 2)
    strcpy(temp2, "0f");

  k = (temp2[0] - '0');

  switch (temp2[1])
  {
  case ('f'):
    if ((k < 11) && (k > -1))
    {
      for (xloc = pl->obj->x - k; xloc <= (signed) pl->obj->x + k; xloc++)
        for (yloc = pl->obj->y - k; yloc <= (signed) pl->obj->y + k; yloc++)
          if (IN_MAP(pl->obj->mapin, xloc, yloc))
            for (tmp = TOP_MAP_OBJ(pl->obj->mapin, xloc, yloc, pl->obj->z);
                 tmp != NULL; tmp = next)
            {
              next = tmp->below;
              if (!IS_PLAYER(tmp))
              {
                remove_ob(tmp);
                free_object(tmp);
              }
            }
    }
    break;
  default:
    write_text(pl, " that macro has not been implemented");
    break;
  }
}

void
inventory_set_val(playerstruct * pl)
{
  char            temp1[15],
                  temp2[15],
                  temp3[15];
  object         *tmp;
  char            tempa[255];

  if (sscanf(pl->parsing, "%s %s %s %[^\n]", temp1, temp2, temp3, tempa) < 4)
  {
    write_text(pl, "insufficient arguments");
    return;
  }
  for (tmp = pl->obj->holding; tmp != NULL && strcmp(tmp->name, temp2);
       tmp = tmp->below);
  if (tmp == NULL)
  {
    write_text(pl, "no such object in inventory");
    return;
  }
  set_obj_val(pl, tmp, temp3, tempa);
}

void
world_set_val(playerstruct * pl)
{
  char            temp1[15],
                  temp2[15],
                  temp3[15];
  object         *tmp;
  char            tempa[255];

  if (sscanf(pl->parsing, "%s %s %s %[^\n]", temp1, temp2, temp3, tempa) < 4)
  {
    write_text(pl, "insufficient arguments");
    return;
  }
  for (tmp = TOP_MAP_OBJ(pl->obj->mapin, pl->obj->x, pl->obj->y, pl->obj->z);
       tmp != NULL && strcmp(tmp->name, temp2); tmp = tmp->below);

  if (tmp == NULL)
  {
    write_text(pl, "no such object in map stack");
    return;
  }
  set_obj_val(pl, tmp, temp3, tempa);
}

void
enter_dm_mode(playerstruct * pl)
{
  char            temp1[15],
                  temp2[200];
  if (dm_mode)
  {
    write_text(pl, "already in strictly dm_mode");
    return;
  }
  if (PL_HAS_DM_POWER(pl))
  {
    write_text(pl, "exiting dm mode");
    CLEAR_PL_HAS_DM_POWER(pl);
    CLEAR_PHASED(pl->obj);
    return;
  }
  if (sscanf(pl->parsing, "%s %15s", temp1, temp2) < 2)
    write_text(pl, "no password ... no dm power!");
  else if (strcmp(DM_PASSWORD, temp2))
    write_text(pl, "no dice bud....");
  else
  {
    SET_PL_HAS_DM_POWER(pl);
    SET_PHASED(pl->obj);
  }
}

int
hash_name(char *name)
{
  int             hashnum = 0,
                  cur = 0;
  for (cur = 0; (name[cur] != '\0'); cur++)
    hashnum += hashnum * 3 + name[cur] * 7;
  hashnum = hashnum % 255;
  return hashnum;
}

int
hash_name2(char *name)
{
  int             hashnum = 0,
                  cur = 0;
  for (cur = 0; (name[cur] != '\0'); cur++)
    hashnum += hashnum * 7 + name[cur] * 3;
  return hashnum;
}

void
hash_commands(void)
{
  command        *tmp,
                 *current = NULL;
  int             counter,
                  hash1;
                  
  /* intialize the hash table */
  for (counter = 0 ; counter < COMMAND_HASH_SIZE; counter ++)
    commlist[counter] = NULL;
 
  for (counter = 0; counter < COMMAND_MAX; counter++)
  {
    tmp = (command *) malloc(sizeof(command));
    tmp->specs = comm_specs[counter];
    hash1 = name_hash(comm_specs[counter].name,COMMAND_HASH_SIZE);
    tmp->rehash = name_rehash(comm_specs[counter].name,COMMAND_HASH_SIZE);
    if (commlist[hash1] == NULL)
    {
      commlist[hash1] = tmp;
      tmp->next = NULL;
    }
    else
    {
      for (current = commlist[hash1]; current->next != NULL;
           current = current->next);
      current->next = tmp;
      current->next->next = NULL;
    }
  }
}
