/* Game_Struct.h 
This is the structure header file. It contains all the structures that will be used throughout the game. It is essential to understand the structures to
understand the game. The structure imposes a pseudo-object oriented environ
that may not be too wieldly, but is more flexible than would C++ allow. */

typedef enum
{
	move_f,
	move_fr,
	move_fl,
	move_r,
	move_l,
	move_b,
	move_bl,
	move_br,
	apply_f,
	apply_fr,
	apply_fl,
	apply_r,
	apply_l,
	apply_b,
	apply_bl,
	apply_br,
	turn_left,
	turn_right,
	shift_up,
	shift_down,
	shift_right,
	shift_left,
 	pickup,
	drop,
	equip,
	repeat,
	interact_mode,
	interact_with,
	use_mode,
	use_obj,
	container,
	query,
	parser,
	aggressive
}		key_press_range;
	
typedef enum
{
	noclass,
  mystic,
  disciple,
  ninja,
  brute,
  warrior,
  treach,
  sniper,
  NUM_CLASS
}               class_range;

typedef enum
{
  torso,
  head,
  legs,
  neck,
  finger
} armour_range;

typedef enum
{
  user,
  enemy,
  george,
  none,
  tele1,
  autotele,
  melee_weap,
  no_key_door,
  key_door,
  key,
  emp1,
  emp2,
  emp3,
  emp4,
  emp5,
  emp6,
  emp7,
  emp8,
  emp9,
  emp10,
  emp11,
  emp12,
  emp13,
  emp14,
  NUM_HANDLES
}               handle_range;

typedef enum
{
  nobonus,
  str,
  sma,
  wis,
  dex,
  con,
  luck,
  looks,
  cha,
  speed,
  NUM_BONUS
}               bonus;

typedef enum
{
  human,
  machine,
  tripod,
  hulk,
  phaeton,
  NUM_RACES
} race_range;

typedef enum
{
  physical,
  mental,
  electric,
  heat,
  cold,
  acid,
  radiation,
  poison
}               damage_range;

/***********************************/
/* range of requests to objects    */
/***********************************/
typedef enum
{
  turn,
  recieve_button,
  recieve_keyboard,
  recieve_attack,
  interact,
  use,
  channel,
  NUM_REQUESTS
}               request;

/***********************************/
/* range of alignments enumeration */
/***********************************/
typedef enum
{
  evil, horrible, bad, naughty,
  nuetral,
  nice, good, wonderful, saintly,
  NUM_ALIGN
}               align;

/***********************************/
/* range of locations in the map   */
/* volumes                         */
/***********************************/
typedef enum
{
  low, middle, high
}               locations;

/***********************************/
/* range of objects enumeration    */
/***********************************/
typedef enum
{
  null,              /* for object links =0              */
  player,            /* special type for players  =1     */
  animate,           /* monsters and things that move =2 */
  inanimate,         /* that which cannot move on its own */
  wall,              /* walls                            */
  floor,             /* floors                           */
  linker,            /* operators (buttons,levers)       */
  door,              /* doors                            */
  teleport,          /* teleporters (is this neccesary?) */
  armour,
  spell,
  weapon,
  ability,
  bad_obj,
  NUM_TYPES          /* dummy variable for counting       */
}               object_type;

/***********************************/
/* statistics for players and for  */
/* animate objects                 */
/***********************************/
typedef struct
{
  int             alignment;

  unsigned int    exper;         /* experience points                 */
  unsigned char   level,         /* what level you are                */
                  str,           /* strength 1 unit = lift 10 pounds  */
                  sma,           /* intelligence 1 unit = 10 iq       */
                  wis,           /* wisdom 1 unit = ?                 */
                  dex,           /* dexterity 1 unit = ?              */
                  con,           /* constitution 1 unit = ?           */
                  luck,          /* luck 1 unit = bonuses             */
                  looks,         /* just how ugly you are             */
                  cha,           /* your personality 20 = ?           */
                  armour,
                  accuracy;
                  
  int             food,          /* food in your stomach              */
                  water,         /* water in your body                */
                  damage;        /* how much damage you do            */
 }               statvars;

/*********************************/
/* physical and mental attribute */
/* structure data                */
/*********************************/
typedef struct
{
  short int       phys,
                  physmax,       /* physical points                   */
                  mental,
                  mentmax;       /* mental points for spells          */
  int             weight,
                  bulk,
                  maxbulk;
}               att;

/*********************************/
/* for the various bitmaps that  */
/* have to be place for an object*/
/*********************************/
typedef struct
{
  unsigned short  picture[4];
  struct pictstruct *next;       /* animation???? */
}               pictstruct;

/*********************************/
/* for objects with messages     */
/*********************************/
typedef struct msg
{
  char           *text;
  struct msg     *next;
}               msg;

typedef struct channel_entry
{
  struct object         *obj;
  struct channel_entry  *prev,
                        *next;
} channel_entry;

typedef struct channel_head
{
  char name[15];
  int  rehash,
       refs;
  channel_entry *entries;
  struct channel_head *next;
} channel_head;

typedef struct channel_ref
{
  channel_head      *head;
  struct channel_entry     *entry;
  struct channel_ref *next;
} channel_ref;

typedef struct
{
  unsigned int    normal:8;
  unsigned int    input:4;
  unsigned int 	  aux:4;
  unsigned int    extended:16;
}               statenums;

/*********************************/
/* status bits for objects       */
/*********************************/
typedef struct obbits
{
  unsigned int    phased:1;      /* pass thru walls?       */
  unsigned int    blocks_view:1; /* can see past?          */
  unsigned int    blocks_move:1; /* can move past          */
  unsigned int    player_view:1; /* can be seen by player? */
  unsigned int    removed:1;     /* is not in a map?       */
  unsigned int    freed:1;       /* has been freed?        */
  unsigned int    pl_visible:1;  /* is visible to players? */
  unsigned int    invisible:1;   /* invisible object?      */
  unsigned int    can_drop:1;    /* can drop this object?  */
  unsigned int    can_take:1;    /* can take this object?  */
  unsigned int    is_equipped:1; /* is this object equipped*/
  unsigned int    savable:1;     /* is this object savable?*/
  unsigned int		has_commlink:1;/* obj on commlink list?  */
  unsigned int    needs_state:1; /* save state w/object?   */
  /* many more to come.... */
}               statbit;

/*********************************/
/*status bits for players        */
/*********************************/
typedef struct plbits
{
  unsigned int    dm_power:1;    /* can perform dm actions?    */
  unsigned int    text_changed:1;/* has text window changed?   */
  unsigned int    stats_changed:1;  /* has stats window changed?  */
  unsigned int    view_changed:1;/* has view window changed?   */
  unsigned int    items_changed:1;  /* has items window changed?  */
  unsigned int    coms_changed:1;
  unsigned int    view_left:1;
  unsigned int    parse_changed:1;
  unsigned int    has_moved:1;
  unsigned int    has_attacked:1;
  unsigned int    has_used;
  unsigned int    charges:1;
  unsigned int    has_exited;
  unsigned int    noresources:1;
  /* many more to come... */
}               plbits;

/*********************************/
/* composition of an object      */
/*********************************/
typedef struct comp
{
  unsigned int    wood:1;
  unsigned int    hardwood:1;
  unsigned int    plastic:1;
  unsigned int    rubber:1;
  unsigned int    glass:1;
  unsigned int    paper:1;
  unsigned int    mineral:1;
  unsigned int    metal:1;
  unsigned int    organix:1;
  unsigned int    energy:1;
  unsigned int    ether:1;
  unsigned int    synthetics:1;
}               comp;

/*********************************/
/* resistances of an object      */
/*********************************/
typedef struct intrinsics
{
  int             heat:4;
  int             cold:4;
  int             electric:4;
  int             physical:4;
  int             mental:4;
  int             poison:4;
  int             radiation:4;
  int             acid:4;
 }               intrinsics;

/*********************************/
/* usage allowances of an object */
/*********************************/
typedef struct
{
  unsigned int  	mystic:1;
  unsigned int  	disciple:1;
  unsigned int  	ninja:1;
  unsigned int  	brute:1;
  unsigned int  	warrior:1;
  unsigned int  	treach:1;
  unsigned int  	sniper:1;
  unsigned int		none:1;
}		use_status;
/*********************************/
/* for pointer to worn objects   */
/*********************************/
typedef struct itemstruct
{
  struct object  *held,
                 *torso,
                 *legs,
                 *head,
                 *right_finger,
                 *left_finger,
                 *neck;
}               itemstruct;

typedef struct movestruct
{
  unsigned int    attack:4;
  unsigned int    normal:4;
  unsigned int    status:8;
}               movestruct;

typedef struct queststruct
{
  unsigned int    quest_num1:4;
  unsigned int    quest_num2:4;
  unsigned int    action1:4;
  unsigned int    action2:4;
}               queststruct;


typedef struct bonusstruct
{
  char           type;
  char           amount;
}               bonusstruct;

/***********************************/
/* data components for all types   */
/***********************************/
typedef struct
{
  struct queststruct quest;
  struct itemstruct items;
  statvars        stats;
  struct playerstruct *pldata;

}               playerd;

typedef struct
{
  struct queststruct quest;
  struct itemstruct items;
  statvars        stats;
  struct movestruct move;
  struct object *interest;
}               animd;

typedef struct
{
  queststruct     quest;
}               inanimd;

typedef struct
{
  queststruct     quest;
  char            mapto[NAMELENGTH];
  unsigned char   x,
                  y,
                  z;
}               teled;

typedef struct
{
  queststruct     quest;
  bonusstruct     bonuses[4];
  char            armour,
                  accuracy;
}               armourd;

typedef struct
{
  queststruct     quest;
  bonusstruct     bonuses[4];
  char            damage,
                  accuracy;
}               weapond;

typedef struct
{
  queststruct     quest;
}               walld;

typedef struct
{
  queststruct     quest;
  char            state;
  int             remaining;
  int		  data;
}               spelld;

typedef struct
{
  char              percentage;
  int               state;
} abilityd;

typedef struct
{
  queststruct     quest;
}               floord;

typedef struct
{
  queststruct     quest;
}               linkerd;

typedef struct
{
  queststruct     quest;
  char		 closed_pic;
  char            keynum;
}               doord;

typedef struct
{
  queststruct     quest;
}               emptyd;

/*********************************/
/* the actual object structure!  */
/* a union of the above structs  */
/* based on type of object       */
/*********************************/
typedef struct object
{
  int             type;
  unsigned char   subtype,
                  facing,
                  speed,
                  color;
  char            handler;
  char            *name;
  int             speed_left;
  use_status      usage;
  unsigned char   x,
                  y,
                  z;

  short int       picnum;

  unsigned int    references,
                  count;
  msg            *message;

  channel_ref    *channels;                  
  struct map     *mapin;

  struct object  *above,
                 *below,
                 *next,
                 *link,
                 *holding,
                 *owner,
                 *gen_ob;
  
  struct comp     material;
  struct intrinsics resist;
  struct obbits   status;
  att             attr;
  statenums       state;

  union data
  {
    playerd        *plcomp;
    inanimd        *incomp;
    animd          *ancomp;
    floord         *flcomp;
    walld          *wacomp;
    spelld         *spcomp;
    armourd        *arcomp;
    weapond        *wecomp;
    linkerd        *licomp;
    doord          *docomp;
    teled          *tecomp;
    abilityd       *abcomp;
    emptyd         *emcomp;
  }               data;

}               object;

/* structure holding data we need in order to draw in windows/pixmaps */
typedef struct
{
  Drawable        window;
  GC              itembg_gc,
                  itemfg_gc,
                  itemxor_gc,
                  text_gc;
  XSizeHints     *size_hint;
}               drawdata;

typedef struct keystruct
{
	KeySym	      keysym;
	int	      function; /* enumerated of keyrange in use */
	struct keystruct *next;
}keystruct;
 
typedef struct playerstruct
{
  int             questcomp,     /* quest completed bits            */
                  queststart,    /* quest started bits              */
                  questreward;   /* recieved reward for quest bits  */
  
  char            keystate,
                  buttonstate,
                  itemmode,
                  dir,
                  exited;
                  
  char						class;
  char            levels[NUM_CLASS];  /* levels per class */

  struct plbits   plstatus;      /* player specific status bits      */
  struct playerstruct *next;     /* if in a group this point to next */
                                 /* player in group                  */

  char            display_name[128];
  Display        *display;       /* what display the player is on    */
  Window          root;          /* display root                     */
  Pixmap          viewpic;

  unsigned long   blackpix,      /* white color and black color     */
                  whitepix;

  int             display_depth,
                  color;

  GC              bw_gc;
  
  Pixmap          masks[NUM_PICS];
  GC              maskgcs[NUM_PICS];

  unsigned long   pixels[34];    /* which pixels we allocated       */

  drawdata        main,          /* window stuff as defined above   */
                  view,
                  text,
                  item,
                  stats,
                  controls,
                  viewpix,
                  itempix;
                  
  keystruct 			*keys;

  Cursor          pointers[CURSORMAX];  /* array of screen cursors */

  /* text maintenance variables */
  short int       tx,
                  ty,
                  th,        /* hieght of font being used          */
                  tc,        /* current line of text               */
                  tw,        /* width of font being used           */
                  tm,        /* maximum line number                */
                  ts,        /* starting line to draw from         */
                  sc,        /* whether or not text is scrolling   */
                  px;

  /* pointers to the fonts that are neccessary for drawing */
  XFontStruct    *itemfont,
                 *cursors,
                 *textfont;

  char            info[80][80],
                  parsing[255];

  object         *pets;        /* pointer to player pets            */
  object         *obj;         /* the object which is the player    */
  object         *top_held;      /* the topmost item in the inventory */
  object         *key_object,
                 *button_object;

  /* function pointers to interaction control functions  */
  void            (*key_control) (),
                  (*button_control) (),
                  (*view_control) ();

  /* data for psuedo-line of sight algorithm             */
  int             in_sight[7][7],
                  lowx,
                  lowy,
                  hix,
                  hiy;
  
  char            save_map[20];
  int             save_x,
                  save_y,
                  save_z;
}               playerstruct;

typedef struct
{
  object         *top,
                 *bottom;
}               stack;

typedef struct square
{
  stack           vol[3];        /* the volume of pointers to  */
  /* objects [low] = lowest     */
  object         *floor;
}               square;

typedef struct message_list
{
	int 		que_num;
	object *obs[20];
	struct message_list *next;
}message_list;

typedef struct map
{
  square          *tiles;
  char            name[NAMELENGTH];
  int             level,         /* difficulty of map            */
                  boundx,        /* maximum x bound              */
                  boundy,        /* maximum y bound              */
                  refs,          /* references in latest round */
                  idle;          /* how long refs = 0 in rounds */
  struct map     *next;
  
  message_list 	 *messages;   /* lists so that objects in same */      				             /* map may 'communicate'         */
} map;


typedef struct
{
  char            name[NAMELENGTH];
  int             status;
  void            (*callback) ();
}               command_spec;
typedef struct command   /* hashed of course */
{
  command_spec    specs;
  int             rehash;
  struct command *next;
}               command;

typedef struct hashed_object
{
  object         *obj;
  struct hashed_object *next;
  int             rehash;
}               hashed_object;

typedef struct hashed_variable
{
  int             rehash;
  void            (*callback) ();
  struct hashed_variable *next;
}               hashed_variable;

typedef struct hashed_name
{
  int                 references,
                      rehash;
  char                name[32];
  struct hashed_name *next;
}               hashed_name;
