/*********************************************************/
/* file: main.c                                          */
/*                                                       */
/* Copyright (C) 1994 Karl A Holland                     */
/*                                                       */
/* author: Karl A Holland                                */
/*                                                       */
/* created:   1 april 93                                 */
/*                                                       */
/* modified:  7 february 94                              */
/*                                                       */
/* contributing authors:                                 */
/*  if you add place your name here or below             */
/*                                                       */
/* description:                                          */
/*  main.c is the main program encapsulation file        */
/* it contains subroutines to start the program and to   */
/* acutually run the program                             */
/*                                                       */
/* LICENSING:                                            */
/*   a full license  is provided in the file LICENSE     */
/* which should have been included with this program when*/
/* it was copied. It may be found in the top directory   */
/* of any NORMAL distribution, named LICENSE             */
/*********************************************************/
#include <standard.h>
#include <signal.h>

void
write_start_text(playerstruct * pl)
{
	FILE           *startfile;
	char            tempbuf[255];
	sprintf(tempbuf, "Welcome to %s, version #%s", TITLE, VERSION);
	write_text(pl, tempbuf);
	write_text(pl,"copyright (c) 1993 Karl Andre' Holland");
	write_text(pl,"(kholland@cs.cornell.edu)");
	sprintf(tempbuf, "[compilation date:%s time:%s]", __DATE__, __TIME__);
	write_text(pl, tempbuf);
	strcpy(tempbuf, LIBDIR);
	strcat(tempbuf,"/");
	strcat(tempbuf, STARTFILE);
	if ((startfile = fopen(tempbuf, "r")))
		while (fscanf(startfile, "%[^\n]", tempbuf) != EOF)
		{
			write_text(pl, tempbuf);
			fscanf(startfile, "%[\n]", tempbuf);
		}
	 else
		fprintf(stderr, "startfile error\n");
}

void
init_player(playerstruct * pl, char *disp_name)
{
  char *name_res,
        tmpbuf[128];
  
  if (disp_name != NULL)
    strcpy(pl->display_name, disp_name);
  else
    pl->display_name[0] = '\0';
  
  make_player_window(pl);
  pl->tc = pl->tx = pl->ts = pl->sc = pl->px = 0;
  pl->th = pl->ty =
    pl->textfont->max_bounds.ascent +
      pl->textfont->max_bounds.descent;
  pl->ty--;
  pl->tw = TEXT_W / pl->textfont->max_bounds.width;
  pl->tm = (TEXT_H / pl->th) - 4;
  pl->keystate = KEY_COMM_STATE;
  pl->button_control = (&get_simple_buttonpress);
  pl->view_control = (&render2);
  pl->key_object = pl->obj;
  pl->button_object = pl->obj;
  pl->buttonstate = 1;
  write_start_text(pl);
  AN_OB_ITEM(pl->obj).held = NULL;
  if ((name_res = XGetDefault(pl->display,RES_NAME,"name")) != NULL)
    {
      pl->obj->state.normal = 3;
      SET_NAME(pl->obj,name_res);
      pl->key_control = (&get_single_key);
      sprintf (tmpbuf,"Name '%s' Retrieved From X Default File",name_res);
      write_text(pl,tmpbuf);
    }
  else
    {
      pl->obj->state.normal = 0;
      pl->key_control = (&get_line);
      write_text (pl,"Please type your name below....");
    }
  pl->itemmode = 1;
  SET_PL_VIEW_LEFT(pl);
  CLEAR_PL_ITEM_CHANGED(pl);
  get_player_keys(pl);
  XFlush(pl->display);
}


void
init_game()
{
  int             count;
  

  fprintf(stderr, "Initiating %s version #%s...\n", TITLE, VERSION);
  
  free_obs = NULL;
  pending_free = NULL;
  list = NULL;
  list_end = NULL;
  players = NULL;
  playerend = NULL;
  maps = NULL;
  
  done = 0;
  
  for (count = 0;count<NAME_HASH_SIZE;count++)
    names[count]=NULL;

  for (count = 0;count < CHANNEL_HASH_SIZE; count ++)
    channel_table[count] = NULL;
  
  tmax = TDELAY;
  
  cycle = 0;
  
  errnum = quit_err = 0;
	
  tmppath = set_up_tmpdir();
  
  XSetErrorHandler(&blowup);

  read_obj_data();
  hash_commands();
  read_pic_data();
}

void
parse_args(int argc, char *argv[])
{
  int       count,numplayers = 0;
  object   *tmp;
  char     tname[17];
  for (count = 1; count < argc; count++)
    {
      if (argv[count][0] == '-')
	{
	  if (!strcmp(argv[count], "-dm"))
	    dm_mode = 1;
	} 
      else
	{
	  tmp = readplayer(TESTFILE);
	  OB_PLDATA(tmp)->obj = tmp;
	  strcpy(tname, argv[count]);
	  strcat(tname, ":0");
	  SET_NAME(tmp, argv[count]);
	  init_player(OB_PLDATA(tmp), tname);
	  send_to_map(tmp,"startmap",9,9+numplayers,0);
	  numplayers++;
	}
    }
  if (!numplayers)
    {
      fprintf(stderr, "%s: insufficient arguments to %s\n", TITLE, TITLE);
      fprintf(stderr, "%s: assuming standard display....\n", TITLE);
      tmp = readplayer(TESTFILE);
      OB_PLDATA(tmp)->obj = tmp;
      init_player(OB_PLDATA(tmp), (char *) NULL);
      send_to_map(tmp,"startmap",9,9,0);
    }
  
  /* set all players to dm status if specified on the command line */ 
  if (dm_mode)
    for (tmp = list;list != NULL && IS_PLAYER(list);list = list->next)
      {
	SET_PL_HAS_DM_POWER (OB_PLDATA(tmp));
	SET_PHASED(tmp);
      }
  
  tmax = TDELAY;
  
  fprintf(stderr, "%s: World build complete.\n", TITLE);
  fprintf(stderr, "%s: Happy Hunting......\n", TITLE);
}


void
cleanup()
{
  object         *tmp,*next;
  char 					comm[255];

  if (++quit_err > MAX_QUIT_ERRORS)
    {
      fprintf (stderr,"recursive error caught!!!!! \n exiting w/o cleanup\n");
      exit (-1);
    }

  fprintf (stderr,"cleaning up....\n");
  
  if (tmppath != NULL && strcmp(tmppath,""))
  {
    sprintf (comm,"rm -f %s/*",tmppath);
    system (comm);
  }
  
  for (tmp = list; tmp != NULL; tmp = next)
    {
      
      next = tmp->next;
      
      switch (tmp->type)
	{	  
	case (player): 
	  free_player(OB_PLDATA(tmp));
	default:
	  free(tmp->data.plcomp);
	  free(tmp);
	  break;
	}
    }

      
      rmdir(tmppath);
}

void
handle_hup()
{
  fprintf (stderr,"recieved SIGHUP\n");
  cleanup();
  exit (-1);
}

void
handle_quit()
{
  fprintf (stderr,"recieved SIGQUIT\n");
  cleanup();
  exit (0);
}

void
handle_bus()
{
 fprintf (stderr,"recieved SIGBUS\n");
  cleanup();
  exit (0);
}

void
handle_segfault()
{
  fprintf (stderr,"recieved SIGSEGV\n");
  cleanup();
  exit (0);
}

void
handle_int()
{
  fprintf (stderr,"recieved SIGINT\n");
  cleanup();
  exit(0);
}

void
set_sigs()
{
  signal (SIGHUP,handle_hup);
  signal (SIGQUIT,handle_quit);
  signal (SIGBUS,handle_bus);
  signal (SIGSEGV,handle_segfault);
  signal (SIGINT,handle_int);
}


void 
player_cleanup(playerstruct * pl)
{
	if (!PL_HAS_EXITED(pl))
	{
	set_sight(pl);
	set_pointer(pl);
	pl->view_control(pl);
	refresh_view(pl);
	if (PL_TEXT_WRITTEN(pl))
		refresh_text(pl);
	if (PL_ITEM_CHANGED(pl))
		refresh_item(pl);
	if (PL_COMS_CHANGED(pl))
		refresh_commands(pl);
	if (PL_STATS_CHANGED(pl))
		refresh_stats(pl);
  XFlush(pl->display);
  }else
  {
  	remove_ob (pl->obj);
  	free_player(pl);
  }
  
}

void
live_cleanup(object * ob)
{
  if (OB_ATTR(ob).phys < OB_ATTR(ob).physmax)
    {
      OB_ATTR(ob).phys += OB_STATS(ob).con / 10;
      if (OB_ATTR(ob).phys > OB_ATTR(ob).physmax)
	OB_ATTR(ob).phys = OB_ATTR(ob).physmax;
      if (IS_PLAYER(ob))
	SET_PL_STATS_CHANGED(OB_PLDATA(ob));
    }
  if (OB_ATTR(ob).mental < OB_ATTR(ob).mentmax)
    {
      OB_ATTR(ob).mental += OB_STATS(ob).sma / 10;
      if (OB_ATTR(ob).mental > OB_ATTR(ob).mentmax)
	OB_ATTR(ob).mental = OB_ATTR(ob).mentmax;
      if (IS_PLAYER(ob))
	SET_PL_STATS_CHANGED(OB_PLDATA(ob));
    }
}


void
main_loop()
{
  object         *current,
  *previous,
  *next;
  int		num_players;
  
  while (!done)
    {
      /* number of players seen so far == 0 */
      num_players = 0;
      
      /*run through entire list of objects */
      for (current = list, previous = NULL; current != NULL;
	   /* previous does not get changed if the object was freed */
	   previous = (OB_IS_FREED(current) ? previous : current),
	   current = next)
	{
	  /* set up next pointer for loop control */
	  next = current->next;
	  
	  /* check to see whether we can enqueue a freed object */
	  if (OB_IS_FREED(current))
	    /*
	     * BIG WARNING : if you unlink an object, its
	     * pointesr are NO LONGER VALID!!!!!!!!!!
	     */
	    unlink_object(previous, current);
	  else
	    {
	      current->speed_left += current->speed;
	      for (; current->speed_left >= SPEED_RESOLUTION;
		   current->speed_left -= SPEED_RESOLUTION)
		SENDRQ(current, turn) (current);
	      if (IS_ALIVE(current) && !(cycle%8))
		live_cleanup(current);
	      if (IS_PLAYER(current))
		{
		  num_players++;
		  (current->mapin->refs)++;
		  player_cleanup(OB_PLDATA(current));
		}
	    }
	  
	}
      /* check free objects for memory maintenance */
      maintain_free();
      
      /* page out old maps to tmp device */
		maintain_maps();
      cycle ++;
      
      /* sleep this process to reduce cpu load and */
      /* increase interface consistency		   */
      usleep(tmax);
      
      /* if no more players, then we're done       */
      if (!num_players)
	done = 1;
    }
}

void
main_loopdm()
{
  object         *current,
  *previous,
  *next;
  int						num_players;
  
  while (!done)
    {
      /* reset number of players seen so far to 0 */
      num_players = 0;
      
      /*run through entire list of objects */
      for (current = players, previous = NULL; current != NULL;
	   previous = (OB_IS_FREED(current) ? previous : current),
	   current = next)
	{
	  /* set up next pointer for loop control */
	  next = current->next;
	  if (OB_IS_FREED(current))
	    
	    /*
	     * BIG WARNING : if you unlink an object, its
	     * pointesr are NO LONGER VALID!!!!!!!!!!
	     */
	    unlink_object(previous, current);
	  else
	    {
	      
	      current->speed_left += current->speed;
	      for (; current->speed_left >= SPEED_RESOLUTION;
		   current->speed_left -= SPEED_RESOLUTION)
		if (IS_PLAYER(current))
		  {
		    SENDRQ(current, turn) (current);
		    current->mapin->refs++;
		  }

	      if (IS_PLAYER(current))
		{
		  player_cleanup(OB_PLDATA(current));
		  num_players++;
		}
	    }
	}
      
      /* check free objects for memory maintenance */
      maintain_free();
      
      /* page out old maps to tmp device */
      maintain_maps();
      
      /* sleep this process to reduce cpu load and */
      /* increase interface consistency	           */
      usleep(tmax);
      
      /* if no more players, then we're done       */
      if (!num_players)
	done = 1;
    }
}
 
int
main(int argc, char *argv[])
{
  set_sigs();
  init_game();
  parse_args (argc,argv);
  if (dm_mode)
    main_loopdm();
  else
    main_loop();
  fprintf (stderr,"cycles: %d\n",cycle);
  cleanup();
  return (0);
}
