/*********************************************************/
/* file: mapfile.c                                       */
/*                                                       */
/* Copyright (C) 1994 Karl A Holland                     */
/*                                                       */
/* author: Karl A Holland                                */
/*                                                       */
/* created:   1 march 93                                 */
/*                                                       */
/* modified:  7 february 94                              */
/*                                                       */
/* contributing authors:                                 */
/*  if you add place your name here or below             */
/*                                                       */
/* description:                                          */
/*  mapfile.c contains subroutines to read and write maps*/
/* to and from disk. It relies heavily on obfile.c       */
/*                                                       */
/* LICENSING:                                            */
/*   a full license  is provided in the file LICENSE     */
/* which should have been included with this program when*/
/* it was copied. It may be found in the top directory   */
/* of any NORMAL distribution, named LICENSE             */
/*********************************************************/
#include <file.h>

char           *
set_up_tmpdir()
{
  int             mypid;
  char           *temp;

  temp = (char *) calloc(255, sizeof(char));
  mypid = getpid();
  sprintf (temp,"/tmp/grudge%d",mypid);
  if (mkdir(temp, 0x1c0))
  {
    fprintf(stderr, "tmp dir error\n");
    exit(2);
  }
  return temp;
}

map
* read_map_data(char *filename)
{
  int             finish,
                  width,
                  length,
                  x,
                  y,
                  z,
                  data = 0;
  object         *tmp_ob;
  char            fname[100];
  map            *tmpmap;
  FILE           *datafile;


  z = finish = 0;
  strcpy(fname, tmppath);
  strcat(fname,"/");
  strcat(fname, filename);

  if (!(datafile = fopen(fname, "r")))
  {
    strcpy(fname, LIBDIR);
    strcat(fname,"/");
    strcat(fname, "maps");
    strcat(fname,"/");
    strcat(fname, filename);
    if (!(datafile = fopen(fname, "r")))
    {
     /* fprintf (stderr,"open to normal directory failed\n"); */
      fprintf(stderr, "could not open map %s\n", filename);
      return (NULL);
    }
  }
  fscanf(datafile, "%d", &data);
  if (data != RD_MAPSTART)
  {
    fprintf(stderr, "bad map file opened.. aborting read");
    fclose(datafile);
    return (NULL);
  }
  fscanf(datafile, "%d", &data);
  if (data != RD_MAPSIZE)
  {
    fprintf(stderr, "reading map without size...aborting");
    fclose(datafile);
    return (NULL);
  }
  fscanf(datafile, "%d %d", &width, &length);
  tmpmap = allocate_map(width, length, filename);
  fscanf(datafile, "%d", &data);
  if (data != RD_MAPLOC)
  {
    fprintf(stderr, "map needs to have x y loc before objects!... aborting");
    deallocate_map(tmpmap);
    fclose(datafile);
    return (NULL);
  }
  fscanf(datafile, "%d %d", &x, &y);
  while (!finish)
  {
    tmp_ob = readobject(datafile, MODE_NEW);
    switch (tmp_ob->type)
    {
    case (BAD):
      fprintf(stderr, "Bad Object in the MapFile:  %s\n",GET_NAME(tmp_ob));
      free_object(tmp_ob);
      break;
    case (MAPLOC):
      fscanf(datafile, "%d %d", &x, &y);
      free_object(tmp_ob);
      break;
    case (FINISHED):
      finish = 1;
      free_object(tmp_ob);
      break;
    case (EOobF):
      fprintf(stderr, "reached EOF inside object...aborting");
      fclose(datafile);
      deallocate_map(tmpmap);
      return (NULL);
      break;
    case (MAPZ):
      fscanf(datafile, "%d", &z);
      free_object(tmp_ob);
      break;
    default:
      insert_ob_in_map(tmpmap, tmp_ob, x, y, z);
      break;
    }
  }
  return (tmpmap);

}

/******************************************/
/* save_map_data                          */
/*    this will save the mapdata in the   */
/* semi-compressed object format on disk  */
/* if mode = 0 (TO_MAP_DIR) it is saved   */
/* to the permanent map directory else if */
/* mode = 1 (TO_TMP_DEV) it is saved to   */
/* the tmp directory for the program for  */
/* paging in later                        */
/******************************************/
int
save_map_data(map * svmap, int mode)
{
  char            fname[100];
  FILE           *datafile;
  int             x,
                  y,
                  z;

  object         *tmp;

  switch (mode)
  {
  case (TO_MAP_DIR):
    strcpy(fname, LIBDIR);
    strcat(fname, "/maps/"); 
    break;
  case (TO_TMP_DEV):
   /* fprintf (stderr,"saving map to tmp device\n");*/
    strcpy(fname, tmppath);
    strcat(fname, "/");
    break;
  default:
    fprintf(stderr, "undefined map save mode %d\n", mode);
    break;
  }

  strcat(fname, svmap->name);

  /*fprintf(stderr, "saving map...%s \n", svmap->name);*/
  if (!(datafile = fopen(fname, "w")))
  {
    fprintf(stderr, "could not open map %s for saving in mode %d\n",
      svmap->name, mode);
    return (0);
  }
  fprintf(datafile, "%d\n", RD_MAPSTART);
  fprintf(datafile, "%d %d %d\n", RD_MAPSIZE, svmap->boundx, svmap->boundy);
  for (z = 0; z < 3; z++)
  {
    for (y = 0; y < svmap->boundy; y++)
      for (x = 0; x < svmap->boundx; x++)
        if (!EMPTY(svmap, x, y, z))
        {
          fprintf(datafile, "%d %d %d\n", RD_MAPLOC, x, y);
          for (tmp = BOT_MAP_OBJ(svmap, x, y, z); tmp != NULL; tmp = tmp->above)
            if ((tmp->type != null) && !IS_PLAYER(tmp))
              write_object(tmp, datafile);
        }
    if (z < 2)
      fprintf(datafile, "%d %d\n", RD_MAPZ, z + 1);
  }

  fprintf(datafile, "%d\n", RD_FINISH);
  fclose(datafile);
 return (1);
}
