/*********************************************************/
/* file: obfile.c                                        */
/*                                                       */
/* Copyright (C) 1994 Karl A Holland                     */
/*                                                       */
/* author: Karl A Holland                                */
/*                                                       */
/* created:   1 march 93                                 */
/*                                                       */
/* modified:  7 february 94                              */
/*                                                       */
/* contributing authors:                                 */
/*  if you add place your name here or below             */
/*                                                       */
/* description:                                          */
/*  obfile.c contains the neccesary subroutines to read  */
/* and write objects to and from disk in an acceptable   */
/* form. it also contains a subroutine that initializes  */
/* the object data base (hash_table) from disk           */
/*                                                       */
/* LICENSING:                                            */
/*   a full license  is provided in the file LICENSE     */
/* which should have been included with this program when*/
/* it was copied. It may be found in the top directory   */
/* of any NORMAL distribution, named LICENSE             */
/*********************************************************/
#include <file.h> 

int
set_up_lockfiles ()
{
  FILE *lockfile;
  char name[128],
       data[256];
  
  strcpy(name,LIBDIR);
  strcat(name,"grudgelock");
  
  /* if no lock file then you are first! */
  if ((lockfile = fopen(name,"r"))==NULL)
    {
    lockfile = fopen (name,"w");
    /*fprintf (lockfile,"%s\n",DisplayString(pl->display));*/
    fclose (lockfile);
    }
  else 
    {
    fscanf(lockfile,"%s",data);
    }
  return 0;
 }

object         *
readobject(FILE * datafile, int mode)
{
  char            strdata[255];
  int             data,
                  data2,
                  data3,
                  done = 0,
                  inside = 0;
  object         *tmp = NULL,
                 *tobj = NULL;
  
  switch (mode)
  {
  case (MODE_GEN):
  case (MODE_NEW):
    break;
  default:
    fprintf(stderr, "tried to get object in unimplemented mode #%d\n", mode);
    exit(1);
  };
  fscanf (datafile,"%d",&data);
  while (!done && data != EOF)
  {
    switch (data)
    {
    case (RD_OBJECT):
      if (!inside)
      {
        switch (mode)
        {
	case (MODE_GEN):
          fscanf(datafile, "%d %d %[^\n]", &data, &data2, strdata);
          tmp = get_blank_object(data);
          SET_TYPE(tmp, data);
          SET_SUBTYPE(tmp, data2);
          SET_NAME(tmp, strdata);
          tmp->gen_ob = NULL;
          break;
        case (MODE_NEW):
          fscanf(datafile, " %[^\n]", strdata);
          tmp = get_general_object(strdata);
          if (tmp == NULL)
          {
            tmp = get_blank_object(inanimate);
            SET_TYPE(tmp, BAD);
          }
          break;
        case (MODE_LINK):
          fscanf(datafile," %[^\n]",strdata);
          tmp = get_blank_object(data);
          SET_NAME(tmp,strdata);
          SET_TYPE(tmp,null);
          SET_SUBTYPE(tmp,0);
          break;
        }
        inside = 1;
      } else
      {
        SET_TYPE(tmp, BAD);
        done = 1;
      }
      break;
    case (RD_HOLDING):
      switch (mode)
      {
      case (MODE_GEN):
        fprintf(stderr, "General Object Holding another Object\n");
        SET_TYPE(tmp, BAD);
        done = 1;
        break;
      case (MODE_NEW):
        add_holding(tmp, readobject(datafile, MODE_NEW));
        if (tmp->holding->type == BAD)
        {
          free_object(tmp->holding);
          done = 1;
          SET_TYPE(tmp, BAD);
        }
      }
      break;
    case (RD_LINK):
      switch (mode)
      {
      case (MODE_LINK):
        fprintf (stderr,"improperly linked object %s",GET_NAME(tmp));
        free_object(readobject (datafile,MODE_LINK));
        break;
      case (MODE_GEN):
        add_link (tmp, readobject(datafile,MODE_GEN));
        break;
      case (MODE_NEW):
        fprintf (stderr,"attempted to read MODE_NEW object with link variable");
        free_object(readobject(datafile,MODE_NEW));
        break;
      }
      break;
    case (RD_NAME):
      switch (mode)
      {
      case (MODE_NEW):
        fscanf(datafile, " %[^\n]", strdata);
        SET_NAME(tmp, strdata);
        break;
      default:
        fprintf(stderr, "tried to set name on general object\n");
        fscanf(datafile, "%*[^\n]");
        break;
      }
      break;
    case (RD_PICNUM):
      switch (mode)
	{
	case (MODE_GEN):
	case (MODE_NEW):
	  for (tobj = tmp; tobj != NULL; tobj = tobj->link)
	    {
	      fscanf(datafile, "%u", &data);
	      SET_PIC(tobj, data);
	    }
	  break;
	case (MODE_LINK):
	  fscanf(datafile, "%u", &data);
	  SET_PIC(tobj, data);
	  break;
	}
      break;
    case (RD_SPEED):
      fscanf(datafile, "%u", &data);
      SET_SPEED(tmp, data);
      break;
    case (RD_STATBITS):
      fscanf(datafile, " %[^\n]", strdata);
      set_char_statbits(strdata, tmp);
      break;
    case (RD_WEIGHT):
      fscanf(datafile, "%d", &data);
      SET_WEIGHT(tmp, data);
      break;
    case (RD_LEVEL):
      fscanf(datafile, "%u", (int *) &data);
      SET_LEVEL(tmp, data);
      break;
    case (RD_STR):
      fscanf(datafile, "%d", (int *) &data);
      SET_STR(tmp, data);
      break;
    case (RD_INT):
      fscanf(datafile, "%d", (int *) &data);
      SET_SMA(tmp, data);
      break;
    case (RD_WIS):
      fscanf(datafile, "%d", (int *) &data);
      SET_WIS(tmp, data);
      break;
    case (RD_DEX):
      fscanf(datafile, "%d", (int *) &data);
      SET_DEX(tmp, data);
      break;
    case (RD_CON):
      fscanf(datafile, "%d", (int *) &data);
      SET_CON(tmp, data);
      break;
    case (RD_LUCK):
      fscanf(datafile, "%d", (int *) &data);
      SET_LUK(tmp, data);
      break;
    case (RD_LOOKS):
      fscanf(datafile, "%d", (int *) &data);
      SET_LOOKS(tmp, data);
      break;
    case (RD_CHA):
      fscanf(datafile, "%d", (int *) &data);
      SET_CHA(tmp, data);
      break;
    case (RD_DAMAGE):
      fscanf(datafile, "%d", &data);
      SET_DAM(tmp, data);
      break;
      
    case (RD_WEAP_DAM):
      if (IS_WEAPON(tmp))
      {
      fscanf(datafile,"%d",&data);
      SET_WEAP_DAM(tmp,data);
      }
      else
      fprintf (stderr,"attempt to set weapon damage in non-weapon object\n");
      break;
    case (RD_END):
      done = 1;
      break;
    case (RD_FINISH):
      tmp = get_blank_object(0);
      tmp->type = FINISHED;
      done = 1;
      break;
    case (RD_MAPLOC):
      tmp = get_blank_object(0);
      tmp->type = MAPLOC;
      done = 1;
      break;
    case (RD_MAPZ):
      tmp = get_blank_object(0);
      tmp->type = MAPZ;
      done = 1;
      break;
    case (RD_EXPER):
      fscanf(datafile, "%d", &(OB_STATS(tmp).exper));
      break;
    case (RD_PHYS):
      fscanf(datafile, "%hd", &(tmp->attr.phys));
      break;
    case (RD_PHYSMAX):
      fscanf(datafile, "%hd", &(tmp->attr.physmax));
      break;
    case (RD_MENTAL):
      fscanf(datafile, "%hd", &(tmp->attr.mental));
      break;
    case (RD_MENTMAX):
      fscanf(datafile, "%hd", &(tmp->attr.mentmax));
      break;
    case (RD_HANDLER):
      fscanf(datafile, "%d", &data);
      tmp->handler=data;
      break;
    case (RD_TELELOC):
      if (IS_TELEPORTER(tmp))
      {
        fscanf(datafile, "%s %d %d %d", (tmp->data.tecomp->mapto)
               ,(int *) &data, (int *) &data2, (int *) &data3);
        tmp->data.tecomp->x = data;
        tmp->data.tecomp->y = data2;
        tmp->data.tecomp->z = data3;
      } else
      {
        fprintf(stderr, "attempt to read teleport data into non-teleporter");
        fscanf(datafile, "%*[^\n]");
      }
      break;
    case (RD_FACING):
      fscanf(datafile, "%d", &data);
      tmp->facing = data;
      break;
    case (RD_BULK):
      fscanf(datafile, "%d", &(OB_ATTR(tmp).bulk));
      break;
    case (RD_MAXBULK):
      fscanf(datafile, "%d", &(OB_ATTR(tmp).maxbulk));
      break;
    case (RD_CURR_BULK):
      fscanf(datafile,"%d", &data);
      break;
    case (RD_CLASS):
      fscanf(datafile,"%d",&data);
      tmp->subtype = data;
      break;
    case (RD_ARMOR_VAL):
      fscanf(datafile,"%d",&data);
      SET_ARMOR_VAL(tmp,data);
      break;
    case (RD_ACC_VAL):
      fscanf(datafile,"%d",&data);
      SET_ARM_ACC(tmp,data);
      break;
    case (RD_ACC):
      fscanf(datafile,"%d",&data);
      SET_ACC(tmp,data);
      break;
    case (RD_MATERIAL):
      fscanf(datafile,"%s",strdata);
      set_obj_material(strdata,tmp);
      break;
    case (RD_RESIST):
      fscanf(datafile,"%s",strdata);
      set_obj_resist(strdata,tmp);
      break;
    case (RD_ARMOUR):
      fscanf(datafile,"%d",&data);
      SET_ARMOUR(tmp,data);
      break;
    case (RD_USAGE):
      fscanf (datafile,"%s",strdata);
      set_obj_usage(strdata,tmp);
      break;
    case (RD_BONUS):
      fscanf (datafile,"%s",strdata);
      set_obj_bonus(strdata,tmp);
      break;
    case (RD_STATE):
      fscanf (datafile,"%d %d %d",&data,&data2,&data3);
      tmp->state.normal = data;
      tmp->state.input = data2;
      tmp->state.aux = data3;
      break;
    case (RD_CLOSEDPIC):
      fscanf(datafile,"%d",&data);
      if(IS_DOOR(tmp))
        tmp->data.docomp->closed_pic = data;
      else
        fprintf (stderr,"attempt to set closed pic in non door structure");
      break;
    case (RD_KEYNUM):
      fscanf (datafile,"%d",&data);
      if(IS_DOOR(tmp))
        tmp->data.docomp->keynum = data;
      else
        fprintf (stderr,"attempt to set keynum in non door structure");
      break;
    case (RD_MSG):
      fscanf (datafile,"%[^~\n]",strdata);
      fscanf (datafile,"%*[~\n]");
      set_message (tmp,strdata);
      break;
    case(RD_CHANNEL):
      fscanf (datafile,"%[^\n]",strdata);
      add_channel_to_obj (tmp,strdata);
      break;
    case(RD_COLOR):
      fscanf (datafile,"%d",&data);
      tmp->color = data;
      break;
    default:
      if (!inside)
      {
        tmp = get_blank_object(0);
        tmp->type = BAD;
      } else
        fprintf(stderr, "unimplemented variable %d in object %s\n", data, tmp->name);
      
    }
    if (!done)
      fscanf(datafile, "%d", &data);

  }

  if (!done)
    tmp->type = EOobF;
  return tmp;
}

void
write_object(object * ob, FILE * datafile)
{
  char            tmpstring[30],
                  tmpstring2[30];

  msg            *tmp_msg;

  object         *tmp,*tmp2;

  channel_ref    *cref;

  fprintf(datafile, "%d %s\n", RD_OBJECT, GET_NAME(ob->gen_ob));
  if (strcmp(GET_NAME(ob), GET_NAME(ob->gen_ob)))
    fprintf(datafile, "%d %s\n", RD_NAME, GET_NAME(ob));
  
  if (GET_SPEED(ob) != GET_SPEED(ob->gen_ob))
    fprintf(datafile, "%d %d\n", RD_SPEED, GET_SPEED(ob));
  if (GET_WEIGHT(ob) != GET_WEIGHT(ob->gen_ob))
    fprintf(datafile, "%d %d\n", RD_WEIGHT, GET_WEIGHT(ob));
  if (ob->facing != ob->gen_ob->facing)
    fprintf(datafile, "%d %d\n", RD_FACING, ob->facing);
  if (GET_MENTAL(ob) != GET_MENTAL(ob->gen_ob))
    fprintf(datafile, "%d %d\n", RD_MENTAL, GET_MENTAL(ob));
  if (GET_MENTMAX(ob) != GET_MENTMAX(ob->gen_ob))
    fprintf(datafile, "%d %d\n", RD_MENTMAX, GET_MENTMAX(ob));
  if (GET_PHYS(ob) != GET_PHYS(ob->gen_ob))
    fprintf(datafile, "%d %d\n", RD_PHYS, GET_PHYS(ob));
  if (GET_PHYSMAX(ob) != GET_PHYSMAX(ob->gen_ob))
    fprintf(datafile, "%d %d\n", RD_PHYSMAX, GET_PHYSMAX(ob));
  if (ob->handler != ob->gen_ob->handler)
    fprintf(datafile, "%d %d\n", RD_HANDLER, ob->handler);
  if (GET_BULK(ob) != GET_BULK(ob->gen_ob))
    fprintf(datafile, "%d %d\n", RD_BULK, GET_BULK(ob));
  if (GET_MAXBULK(ob) != GET_MAXBULK(ob->gen_ob))
    fprintf(datafile, "%d %d\n", RD_MAXBULK, GET_MAXBULK(ob));
  if (ob->color != ob->gen_ob->color)
    fprintf(datafile, "%d %d\n", RD_COLOR,ob->color);
  get_char_statbits(tmpstring, ob);
  get_char_statbits(tmpstring2, ob->gen_ob);
  if (strcmp(tmpstring, tmpstring2))
    fprintf(datafile, "%d %s\n", RD_STATBITS, tmpstring);
    
  get_obj_material(tmpstring,ob);
  get_obj_material(tmpstring2,ob->gen_ob);
  if (strcmp(tmpstring, tmpstring2))
    fprintf(datafile, "%d %s\n", RD_MATERIAL, tmpstring);
  
  get_obj_resist(tmpstring,ob);
  get_obj_resist(tmpstring2,ob->gen_ob);
  if (strcmp(tmpstring, tmpstring2))
    fprintf(datafile, "%d %s\n", RD_RESIST, tmpstring);
  
  get_obj_usage(tmpstring,ob);
  get_obj_usage(tmpstring2,ob->gen_ob);
  if (strcmp(tmpstring, tmpstring2))
    fprintf(datafile, "%d %s\n", RD_USAGE, tmpstring);
  
  if (NEEDS_STATE(ob))
    fprintf(datafile, "%d %d %d %d\n",RD_STATE,
	    (int)ob->state.normal,(int)ob->state.input,(int)ob->state.aux);
    
  if (IS_ALIVE(ob))
  {
    if (GET_LEVEL(ob) != GET_LEVEL(ob->gen_ob))
      fprintf(datafile, "%d %d\n", RD_LEVEL, GET_LEVEL(ob));
    if (GET_EXP(ob) != GET_EXP(ob->gen_ob))
      fprintf(datafile, "%d %d\n", RD_EXPER, GET_EXP(ob));
    if (GET_STR(ob) != GET_STR(ob->gen_ob))
      fprintf(datafile, "%d %d\n", RD_STR, GET_STR(ob));
    if (GET_SMA(ob) != GET_SMA(ob->gen_ob))
      fprintf(datafile, "%d %d\n", RD_INT, GET_SMA(ob));
    if (GET_WIS(ob) != GET_WIS(ob->gen_ob))
      fprintf(datafile, "%d %d\n", RD_WIS, GET_WIS(ob));
    if (GET_DEX(ob) != GET_DEX(ob->gen_ob))
      fprintf(datafile, "%d %d\n", RD_DEX, GET_DEX(ob));
    if (GET_CON(ob) != GET_CON(ob->gen_ob))
      fprintf(datafile, "%d %d\n", RD_CON, GET_CON(ob));
    if (GET_LUK(ob) != GET_LUK(ob->gen_ob))
      fprintf(datafile, "%d %d\n", RD_LUCK, GET_LUK(ob));
    if (GET_CON(ob) != GET_CON(ob->gen_ob))
      fprintf(datafile, "%d %d\n", RD_CON, GET_CON(ob));
    if (GET_LOOKS(ob) != GET_LOOKS(ob->gen_ob))
      fprintf(datafile, "%d %d\n", RD_LOOKS, GET_LOOKS(ob));
    if (GET_CHA(ob) != GET_CHA(ob->gen_ob))
      fprintf(datafile, "%d %d\n", RD_CHA, GET_CHA(ob));
    if (GET_DAM(ob) != GET_DAM(ob->gen_ob))
      fprintf(datafile, "%d %d\n", RD_DAMAGE, GET_DAM(ob));
  }
  if (IS_TELEPORTER(ob))
  {
    if (ob->data.tecomp->mapto[0] != '\0')
      fprintf(datafile, "%d %s %d %d %d\n", RD_TELELOC,
        ob->data.tecomp->mapto, ob->data.tecomp->x, ob->data.tecomp->y,
        ob->data.tecomp->z);
  }
  if (IS_WEAPON(ob))
  {
    if (GET_WEAP_DAM(ob) != GET_WEAP_DAM(ob->gen_ob))
      fprintf (datafile,"%d %d\n",RD_WEAP_DAM,GET_WEAP_DAM(ob));
    if (GET_WEAP_ACC(ob) != GET_WEAP_ACC(ob->gen_ob))
      fprintf (datafile,"%d %d\n",RD_ACC_VAL,GET_ARM_ACC(ob));
  }
  if (IS_DOOR(ob))
  {
    fprintf(datafile,"%d %d\n",RD_KEYNUM,ob->data.docomp->keynum);
    if (ob->data.docomp->closed_pic != ob->gen_ob->data.docomp->closed_pic)
      fprintf(datafile,"%d %d",RD_CLOSEDPIC,ob->data.docomp->closed_pic);
  }       
  if (IS_ARMOUR(ob))
  {
    if (GET_ARMOR_VAL(ob) != GET_ARMOR_VAL(ob->gen_ob))
      fprintf (datafile,"%d %d\n",RD_ARMOR_VAL,GET_ARMOR_VAL(ob));
    if (GET_ARM_ACC(ob) != GET_ARM_ACC(ob->gen_ob))
      fprintf (datafile,"%d %d\n",RD_ACC_VAL,GET_ARM_ACC(ob));
  }
  if (HAS_ITEMS(ob))
  {
    fprintf(datafile, "%d\n", RD_HOLDING);
    for (tmp = ob->holding; tmp != null; tmp = tmp->above)
      write_object(tmp, datafile);
  }
  if (ob->message != NULL)
    {
      for(tmp_msg = ob->message; tmp_msg != NULL; tmp_msg = tmp_msg->next)
	fprintf (datafile,"%d %s~\n",RD_MSG,tmp_msg->text);
    }
  if (ob->channels != NULL)
    {
      for (cref = ob->channels; cref != NULL; cref = cref->next )
	fprintf (datafile,"%d %s\n",RD_CHANNEL,cref->head->name);
    }

  for (tmp = ob,tmp2=ob->gen_ob; tmp != NULL; tmp = tmp->link,tmp2= tmp2->link)
    if ( GET_PIC(tmp) != GET_PIC(tmp2))
      break;
  if (tmp != NULL)
    {
      fprintf(datafile, "%d", RD_PICNUM);
      for (tmp = ob ; tmp!= NULL; tmp = tmp->link)
	fprintf(datafile, " %d\n",GET_PIC(tmp));
      fprintf (datafile,"\n");
    }
  fprintf(datafile, "%d\n", RD_END);
}

int
read_obj_data()
{
  
  FILE           *datafile;
  char            strdata[255];
  object         *tmpob;
 /* hashed_object  *tloc,
                 *curr;*/
  int             hashlock,
                  done = 0,
                  data = 1000;

  fprintf(stderr, "starting object read\n"); 

  /* null the hash table */
  for (hashlock = 0; hashlock < OBJECT_HASH_SIZE; hashlock++)
    obdata[hashlock]= NULL;

  /* construct the path */
  strcpy(strdata, LIBDIR);
  strcat(strdata,"/");
  strcat(strdata, OBJFILE);
  if ((datafile = fopen(strdata, "r")))
  {
    fscanf(datafile, "%d", &data);
    fprintf(stderr, "%d\n", data);
    if (data != RD_START)
    {
      fprintf(stderr, "bad object file \n");
      fclose(datafile);
      return 1;
    } else
    {
      fprintf(stderr, "reading datafile \n");
      while (!done)
      {
        tmpob = readobject(datafile, MODE_GEN);
        switch (tmpob->type)
        {
        case (BAD):
          fprintf(stderr, "bad object in file: %s\n",GET_NAME(tmpob));
          fclose(datafile);
          free_object(tmpob);
          return 1;
          break;
        case (EOobF):
          fprintf(stderr, "reached EOF in object\n");
          free_object(tmpob);
          break;
        case (FINISHED):      
          free_object(tmpob);
          done = 1;
          break;
        default:
	  add_to_hashlist(tmpob);
	  /*
          hashlock = name_hash(tmpob->namem);
          tloc = (hashed_object *) malloc(sizeof(hashed_object));
          tloc->obj = tmpob;
          tloc->rehash = obj_rehash(tmpob->name);
          if (obdata[hashlock] == NULL)
            obdata[hashlock] = tloc;
          else
          {
            for (curr = obdata[hashlock]; curr->next != NULL;
                 curr = curr->next);
            curr->next = tloc;
          }
	  tloc->next = NULL;*/
          break;
        }
      }
      if (!done)
      {
        fprintf(stderr, "reached EOF w/o finish\n");
        fclose(datafile);
        return 1;
      }

      
        
      
    }
    fclose(datafile);
    return 0;
  }
  return 0;
}

void
read_pic_data()
{
  FILE           *datafile;
  char            strdata[255];
  int             index,
                  ul,
                  ur,
                  lr,
                  ll;

  strcpy(strdata, LIBDIR);
  strcat(strdata, "/");
  strcat(strdata, PICFILE);
  if ((datafile = fopen(strdata, "r")) == NULL)
  {
    fprintf(stderr, "bad picture file\n");
    exit(2);
  }
  fscanf(datafile, "%d", &index);
  while (index != -1)
  {
    if (index < 1001)
    {
      fscanf(datafile, "%d %d %d %d %*[^\n]", &ul, &ur, &lr, &ll);
      picnums[index][0] = ul;
      picnums[index][1] = ur;
      picnums[index][2] = lr;
      picnums[index][3] = ll;

      /*
       * fprintf (stderr,"%d %d %d %d %d
       * \n",index,ul,ur,lr,ll);
       */
    } else
    {
      fscanf(datafile, "%*d %*d %*d %*d %*[^\n]");
      fprintf(stderr, "bad pic num #%d\n", index);
    }
    fscanf(datafile, "%d", &index);
  }
  /*fprintf(stderr, "completed read of picture file\n");*/
}
